<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\Controller;

// Prevent direct access
defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\ControllerEvents;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\GetRedirectionAware;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\RegisterControllerTasks;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\ReturnURLAware;
use Akeeba\Component\LoginGuard\Administrator\Controller\Mixin\ReusableModels;
use Akeeba\Component\LoginGuard\Administrator\Event\NotifyActionLog;
use Akeeba\Component\LoginGuard\Administrator\Event\Validate;
use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Model\BackupcodesModel;
use Akeeba\Component\LoginGuard\Administrator\Model\CaptiveModel;
use Akeeba\Component\LoginGuard\Administrator\Model\RemembermeModel;
use Exception;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Controller\BaseController;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use Joomla\CMS\Router\Route;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Input\Input;
use ReflectionObject;
use RuntimeException;

class CaptiveController extends BaseController
{
	use ControllerEvents;
	use GetRedirectionAware;
	use RegisterControllerTasks;
	use ReturnURLAware;
	use ReusableModels;

	public function __construct($config = [], MVCFactoryInterface $factory = null, ?CMSApplication $app = null, ?Input $input = null)
	{
		parent::__construct($config, $factory, $app, $input);

		$this->registerDefaultTask('Captive');
	}

	public function captive($cachable = false, $urlparams = false): void
	{
		$user = $this->app->getIdentity()
			?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);

		$this->setSiteTemplateStyle();

		// Only allow logged in Users
		if ($user->guest)
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// Get the view object
		$viewLayout = $this->input->get('layout', 'default', 'string');
		$view       = $this->getView('Captive', 'html', '', [
			'base_path' => $this->basePath,
			'layout'    => $viewLayout,
		]);

		$view->document = $this->app->getDocument();

		// If we're already logged in go to the site's home page
		if ($this->app->getSession()->get('com_loginguard.mfa_checked', 0) == 1)
		{
			$url = Route::_('index.php?option=com_loginguard&task=methods.display', false);

			$this->setRedirect($url);
		}

		// Pass the model to the view
		/** @var CaptiveModel $model */
		$model = $this->getModel('Captive', 'Administrator');
		$view->setModel($model, true);

		/** @var BackupcodesModel $codesModel */
		$codesModel = $this->getModel('Backupcodes', 'Administrator');
		$view->setModel($codesModel, false);

		try
		{
			// kill all modules on the page
			$model->killAllModules();
		}
		catch (Exception $e)
		{
			// If we can't kill the modules we can still survive.
		}

		// Pass the MFA record ID to the model
		$record_id = $this->input->getInt('record_id', null);
		$model->setState('record_id', $record_id);

		// Validate by Browser ID
		try
		{
			$browserId = $this->getBrowserId();
		}
		catch (Exception $e)
		{
			$browserId = '';
		}

		/** @var RemembermeModel $rememberMeModel */
		$rememberMeModel = $this->getModel('Rememberme', 'Administrator');

		if (!is_null($browserId) && $rememberMeModel->setBrowserId($browserId)->hasValidCookie())
		{
			// Tell the plugins that we successfully applied MFA – used by our User Actions Log plugin.
			MfaHelper::triggerEvent(
				new NotifyActionLog('onComLoginguardCaptiveValidateSuccess', [
					Text::_('COM_LOGINGUARD_LBL_METHOD_BROWSERID'),
				])
			);

			// Flag the user as fully logged in
			$this->app->getSession()->set('com_loginguard.mfa_checked', 1);

			// Get the return URL stored by the plugin in the session
			$return_url = $this->app->getSession()->get('com_loginguard.return_url', '');

			// If the return URL is not set or not inside this site redirect to the site's front page
			if (empty($return_url) || !Uri::isInternal($return_url))
			{
				$return_url = Uri::base();
			}

			$this->setRedirect($return_url);
		}

		// Do not go through $this->display() because it overrides the model, nullifying the whole concept of MVC.
		$view->display();
	}


	/**
	 * Validate the MFA code entered by the user
	 *
	 * @param   bool   $cachable         Can this view be cached
	 * @param   array  $urlparameters    An array of safe url parameters and their variable types, for valid values see
	 *                                   {@link JFilterInput::clean()}.
	 *
	 * @return  void
	 */
	public function validate($cachable = false, $urlparameters = [])
	{
		// CSRF Check
		$this->checkToken($this->input->getMethod());

		$this->setSiteTemplateStyle();

		// Get the MFA parameters from the request
		$record_id  = $this->input->getInt('record_id', null);
		$rememberMe = $this->input->getBool('rememberme', false);
		$code       = $this->input->get('code', null, 'raw');
		/** @var CaptiveModel $model */
		$model = $this->getModel('Captive', 'Administrator');

		// Validate the MFA record
		$model->setState('record_id', $record_id);
		$record = $model->getRecord();

		if (empty($record))
		{
			MfaHelper::triggerEvent(
				new NotifyActionLog('onComLoginguardCaptiveValidateInvalidMethod')
			);

			throw new RuntimeException(Text::_('COM_LOGINGUARD_ERR_INVALID_METHOD'), 500);
		}

		// Validate the code
		$user = $this->app->getIdentity()
			?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);

		$results     = MfaHelper::triggerEvent(new Validate($record, $user, $code));
		$isValidCode = false;

		if ($record->method == 'backupcodes')
		{
			/** @var BackupcodesModel $codesModel */
			$codesModel = $this->getModel('Backupcodes', 'Administrator');
			$results    = [$codesModel->isBackupCode($code, $user)];
			/**
			 * This is required! Do not remove!
			 *
			 * There is a store() call below. It saves the in-memory MFA record to the database. That includes the
			 * options key which contains the configuration of the Method. For backup codes, these are the actual codes
			 * you can use. When we check for a backup code validity we also "burn" it, i.e. we remove it from the
			 * options table and save that to the database. However, this DOES NOT update the $record here. Therefore
			 * the call to saveRecord() would overwrite the database contents with a record that _includes_ the backup
			 * code we had just burned. As a result the single use backup codes end up being multiple use.
			 *
			 * By doing a getRecord() here, right after we have "burned" any correct backup codes, we resolve this
			 * issue. The loaded record will reflect the database contents where the options DO NOT include the code we
			 * just used. Therefore the call to store() will result in the correct database state, i.e. the used backup
			 * code being removed.
			 */
			$record = $model->getRecord();
		}

		if (is_array($results) && !empty($results))
		{
			foreach ($results as $result)
			{
				if ($result)
				{
					$isValidCode = true;

					break;
				}
			}
		}

		if (!$isValidCode)
		{
			// The code is wrong. Display an error and go back.
			$captiveURL = Route::_('index.php?option=com_loginguard&view=captive&record_id=' . $record_id, false);
			$message    = Text::_('COM_LOGINGUARD_ERR_INVALID_CODE');
			$this->setRedirect($captiveURL, $message, 'error');

			MfaHelper::triggerEvent(
				new NotifyActionLog('onComLoginguardCaptiveValidateFailed', [$record->title])
			);

			return;
		}

		// Handle the Remember Me option
		$browserId = $this->getBrowserId();

		if ($rememberMe && !empty($browserId))
		{
			/** @var RemembermeModel $remembermeModel */
			$remembermeModel = $this->getModel('Rememberme', 'Administrator');
			$remembermeModel
				->setBrowserId($browserId)
				->setUsername($user->username)
				->setCookie();
		}

		// Update the Last Used, UA and IP columns
		$jNow = Date::getInstance();

		$record->last_used = $jNow->toSql();
		$record->store();

		// Flag the user as fully logged in
		$session = $this->app->getSession();
		$session->set('com_loginguard.mfa_checked', 1);

		// Get the return URL stored by the plugin in the session
		$return_url = $session->get('com_loginguard.return_url', '');

		// If the return URL is not set or not inside this site redirect to the site's front page
		if (empty($return_url) || !Uri::isInternal($return_url))
		{
			$return_url = Uri::base();
		}

		$this->setRedirect($return_url);

		MfaHelper::triggerEvent(
			new NotifyActionLog('onComLoginguardCaptiveValidateSuccess', [$record->title])
		);
	}

	/**
	 * Get the Browser ID from the session or the request
	 *
	 * Checks if there is a valid request trying to set the browser ID. If so, it's saved in the session and returned.
	 * Otherwise we return whatever browser ID we currently have in the session.
	 *
	 * @return string|null
	 * @throws Exception
	 */
	private function getBrowserId(): ?string
	{
		$session = $this->app->getSession();

		/**
		 * I will only accept a browser ID if it's POSTed with a valid anti-CSRF token and the session flag is set. This
		 * gives me adequate assurances that there's no monkey business going on.
		 */
		try
		{
			$allowedFlag = ($this->input->getMethod() == 'POST') &&
				$session->get('com_loginguard.browserIdCodeLoaded', false);
		}
		catch (Exception $e)
		{
			$allowedFlag = false;
		}

		if ($allowedFlag)
		{
			// IMPORTANT: DO NOT USE checkToken() — IT CAUSES AN INFINITE REDIRECTION!
			$token       = $this->app->getFormToken();
			$allowedFlag = $this->input->post->getInt($token, 0) == 1;
		}

		// Get the browser ID recorded in the session and in the request
		$browserIdSession = $session->get('com_loginguard.browserId', null);
		$browserIdRequest = $this->input->post->getString('browserId', null);

		// Nobody is trying to set a browser ID in the request. Return the browser ID we stored in the session.
		if (!$allowedFlag && is_null($browserIdRequest))
		{
			return $browserIdSession;
		}

		// Attempt to pass a browser ID from a page other than layout=fingerprint. Pretend fingerprinting failed.
		if (!$allowedFlag)
		{
			$browserIdRequest = '';
		}

		// We already have a browser ID in the session and we're given a different one. That's... strange.
		if (!is_null($browserIdSession) && !empty($browserIdRequest) && ($browserIdRequest != $browserIdSession))
		{
			$browserIdRequest = '';
		}

		// Normalize zero, null and empty string to an empty string that means "fingerprinting failed"
		if (empty($browserIdRequest))
		{
			$browserIdRequest = '';
		}

		// Reset the flag to prevent opportunities to override our browser fingerprinting
		$session->set('com_loginguard.browserIdCodeLoaded', false);
		// Save the browser ID in the session
		$session->set('com_loginguard.browserId', $browserIdRequest);

		// Finally, return the browser ID as requested
		return $browserIdRequest;
	}

	/**
	 * Set a specific site template style in the frontend application
	 *
	 */
	private function setSiteTemplateStyle(): void
	{
		$Itemid = $this->input->get('Itemid');

		if (!empty($Itemid))
		{
			return;
		}

		$templateStyle = (int) ComponentHelper::getParams('com_loginguard')->get('captive_template', '');

		if (empty($templateStyle) || !$this->app->isClient('site'))
		{
			return;
		}

		$this->app->input->set('templateStyle', $templateStyle);

		$refApp      = new ReflectionObject($this->app);
		$refTemplate = $refApp->getProperty('template');
		$refTemplate->setAccessible(true);
		$refTemplate->setValue($this->app, null);

		$template = $this->app->getTemplate(true);

		$this->app->set('theme', $template->template);
		$this->app->set('themeParams', $template->params);
	}
}