<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\Model;

// Prevent direct access
defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\DataShape\SetupRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowSetup;
use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Table\TfaTable;
use Exception;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Model\BaseDatabaseModel;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;

/**
 * Multi-factor Authentication Method management model
 */
class MethodModel extends BaseDatabaseModel
{
	/**
	 * List of MFA Methods
	 *
	 * @var   array
	 * @since 1.2.0
	 */
	protected $mfaMethods = null;

	/**
	 * Get the specified MFA Method's record
	 *
	 * @param   string  $method  The Method to retrieve.
	 *
	 * @return  array
	 * @since   1.2.0
	 */
	public function getMethod(string $method): array
	{
		if (!$this->methodExists($method))
		{
			return [
				'name'          => $method,
				'display'       => '',
				'shortinfo'     => '',
				'image'         => '',
				'canDisable'    => true,
				'allowMultiple' => true,
			];
		}

		return $this->mfaMethods[$method];
	}

	/**
	 * Is the specified MFA Method available?
	 *
	 * @param   string  $method  The Method to check.
	 *
	 * @return  bool
	 * @since   1.2.0
	 */
	public function methodExists(string $method): bool
	{
		if (!is_array($this->mfaMethods))
		{
			$this->populateMfaMethods();
		}

		return isset($this->mfaMethods[$method]);
	}

	/**
	 * @param   User|null  $user  The user record. Null to use the currently logged in user.
	 *
	 * @return  array
	 * @throws  Exception
	 *
	 * @since   1.2.0
	 */
	public function getRenderOptions(?User $user = null): SetupRenderOptions
	{
		if (is_null($user))
		{
			$user = Factory::getApplication()->getIdentity() ?: Factory::getUser();
		}

		$renderOptions = new SetupRenderOptions();

		$results = MfaHelper::triggerEvent(new ShowSetup($this->getRecord($user)));

		if (empty($results))
		{
			return $renderOptions;
		}

		foreach ($results as $result)
		{
			if (empty($result))
			{
				continue;
			}

			return $renderOptions->merge($result);
		}

		return $renderOptions;
	}

	/**
	 * Get the specified MFA record. It will return a fake default record when no record ID is specified.
	 *
	 * @param   User|null  $user  The user record. Null to use the currently logged in user.
	 *
	 * @return  TfaTable
	 * @throws  Exception
	 *
	 * @since   1.2.0
	 */
	public function getRecord(User $user = null): TfaTable
	{
		if (is_null($user))
		{
			$user = Factory::getApplication()->getIdentity()
				?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);

		}

		$defaultRecord = $this->getDefaultRecord($user);
		$id            = (int) $this->getState('id', 0);

		if ($id <= 0)
		{
			return $defaultRecord;
		}

		/** @var TfaTable $record */
		$record = $this->getTable('Tfa', 'Administrator');
		$loaded = $record->load([
			'user_id' => $user->id,
			'id'      => $id,
		]);

		if (!$loaded)
		{
			return $defaultRecord;
		}


		if (!$this->methodExists($record->method))
		{
			return $defaultRecord;
		}

		return $record;
	}

	/**
	 * Return the title to use for the page
	 *
	 * @return  string
	 *
	 * @since   1.2.0
	 */
	public function getPageTitle(): string
	{
		$task = $this->getState('task', 'edit');

		return Text::_(sprintf("COM_LOGINGUARD_HEAD_%s_PAGE", $task));
	}

	/**
	 * @param   User|null  $user  The user record. Null to use the current user.
	 *
	 * @return  TfaTable
	 * @throws  Exception
	 *
	 * @since   1.2.0
	 */
	protected function getDefaultRecord(?User $user = null): TfaTable
	{
		if (is_null($user))
		{
			$user = Factory::getApplication()->getIdentity()
				?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);
		}

		$method = $this->getState('method');
		$title  = '';

		if (is_null($this->mfaMethods))
		{
			$this->populateMfaMethods();
		}

		if ($method && isset($this->mfaMethods[$method]))
		{
			$title = $this->mfaMethods[$method]['display'];
		}

		/** @var TfaTable $record */
		$record = $this->getTable('Tfa', 'Administrator');

		$record->bind([
			'id'      => null,
			'user_id' => $user->id,
			'title'   => $title,
			'method'  => $method,
			'default' => 0,
			'options' => [],
		]);

		return $record;
	}

	/**
	 * Populate the list of MFA Methods
	 *
	 * @since   1.2.0
	 */
	private function populateMfaMethods(): void
	{
		$this->mfaMethods = [];
		$mfaMethods       = MfaHelper::getMfaMethods();

		if (empty($mfaMethods))
		{
			return;
		}

		foreach ($mfaMethods as $method)
		{
			$this->mfaMethods[$method['name']] = $method;
		}

		// We also need to add the backup codes Method
		$this->mfaMethods['backupcodes'] = [
			'name'          => 'backupcodes',
			'display'       => Text::_('COM_LOGINGUARD_LBL_BACKUPCODES'),
			'shortinfo'     => Text::_('COM_LOGINGUARD_LBL_BACKUPCODES_DESCRIPTION'),
			'image'         => 'media/com_loginguard/images/emergency.svg',
			'canDisable'    => false,
			'allowMultiple' => false,
		];
	}
}