<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\Service;

use Joomla\CMS\Encrypt\Aes;
use Joomla\CMS\Filesystem\File;
use Joomla\CMS\User\UserHelper;

/**
 * Data encryption service.
 *
 * @since       6.0.0
 */
class Encrypt
{
    /**
     * The encryption engine used by this service
     *
     * @var    Aes
     * @since  6.0.0
     */
    private $aes;

    /**
     * EncryptService constructor.
     *
     * @since   6.0.0
     */
    public function __construct()
    {
        $this->initialize();
    }

    /**
     * Encrypt the plaintext $data and return the ciphertext prefixed by ###AES128###
     *
     * @param string $data The plaintext data
     *
     * @return  string  The ciphertext, prefixed by ###AES128###
     *
     * @since   6.0.0
     */
    public function encrypt(string $data): string
    {
        if (!is_object($this->aes))
        {
            return $data;
        }

	    $this->aes->setPassword($this->getPassword(), false);
        $encrypted = $this->aes->encryptString($data, true);

        return '###AES128###' . $encrypted;
    }

    /**
     * Decrypt the ciphertext, prefixed by ###AES128###, and return the plaintext.
     *
     * @param   string  $data    The ciphertext, prefixed by ###AES128###
     * @param   bool    $legacy  Use legacy key expansion? Use it to decrypt data encrypted with FOF 3.
     *
     * @return  string  The plaintext data
     *
     * @since   6.0.0
     */
    public function decrypt(string $data, bool $legacy = false): string
    {
        if (substr($data, 0, 12) != '###AES128###')
        {
            return $data;
        }

        $data = substr($data, 12);

        if (!is_object($this->aes))
        {
            return $data;
        }

		$this->aes->setPassword($this->getPassword(), $legacy);
        $decrypted = $this->aes->decryptString($data, true, $legacy);

        // Decrypted data is null byte padded. We have to remove the padding before proceeding.
        return rtrim($decrypted, "\0");
    }

    /**
     * Initialize the AES cryptography object
     *
     * @return  void
     * @since   6.0.0
     */
    private function initialize(): void
    {
        if (is_object($this->aes))
        {
            return;
        }

        $password = $this->getPassword();

        if (empty($password))
        {
            return;
        }

        $this->aes = new Aes('cbc');
        $this->aes->setPassword($password);
    }

    /**
     * Returns the path to the secret key file
     *
     * @return  string
     *
     * @since   6.0.0
     */
    private function getPasswordFilePath(): string
    {
        return JPATH_ADMINISTRATOR . '/components/com_loginguard/encrypt_service_key.php';
    }

    /**
     * Get the name of the constant where the secret key is stored. Remember that this is searched first, before a new
     * key file is created. You can define this constant anywhere in your code loaded before the encryption service is
     * first used to prevent a key file being created.
     *
     * @return  string
     *
     * @since   6.0.0
     */
    private function getConstantName(): string
    {
        // Compatible with LoginGuard 4 and earlier
        return 'LOGINGUARD_FOF_ENCRYPT_SERVICE_SECRETKEY';
    }

    /**
     * Returns the password used to encrypt information in the component
     *
     * @return  string
     *
     * @since   6.0.0
     */
    private function getPassword(): string
    {
        $constantName = $this->getConstantName();

        // If we have already read the file just return the key
        if (defined($constantName))
        {
            return constant($constantName);
        }

        // Do I have a secret key file?
        $filePath = $this->getPasswordFilePath();

        // I can't get the path to the file. Cut our losses and assume we can get no key.
        if (empty($filePath))
        {
            define($constantName, '');

            return '';
        }

        // If not, try to create one.
        if (!file_exists($filePath))
        {
            $this->makePasswordFile();
        }

        // We failed to create a new file? Cut our losses and assume we can get no key.
        if (!file_exists($filePath) || !is_readable($filePath))
        {
            define($constantName, '');

            return '';
        }

        // Try to include the key file
        include_once $filePath;

        // The key file contains garbage. Treason! Cut our losses and assume we can get no key.
        if (!defined($constantName))
        {
            define($constantName, '');

            return '';
        }

        // Finally, return the key which was defined in the file (happy path).
        return constant($constantName);
    }

    /**
     * Create a new secret key file using a long, randomly generated password. The password generator uses a crypto-safe
     * pseudorandom number generator (PRNG) to ensure suitability of the password for encrypting data at rest.
     *
     * @return  void
     *
     * @since   6.0.0
     */
    private function makePasswordFile(): void
    {
        // Get the path to the new secret key file.
        $filePath = $this->getPasswordFilePath();

        // I can't get the path to the file. Sorry.
        if (empty($filePath))
        {
            return;
        }

        $secretKey    = UserHelper::genRandomPassword(64);
        $constantName = $this->getConstantName();

        $fileContent = "<?" . 'ph' . "p\n\n";
        $fileContent .= <<< END
defined('_JEXEC') or die;

/**
 * This file is automatically generated. It contains a secret key used for encrypting data by the component. Please do
 * not remove, edit or manually replace this file. It will render your existing encrypted data unreadable forever.
 */
 
define('$constantName', '$secretKey');

END;

        File::write($filePath, $fileContent);
    }
}