<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\LoginGuard\Administrator\View\Methods;

// Prevent direct access
defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\DataShape\MethodDescriptor;
use Akeeba\Component\LoginGuard\Administrator\Event\NotifyActionLog;
use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Model\BackupcodesModel;
use Akeeba\Component\LoginGuard\Administrator\Model\MethodsModel;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\View\HtmlView as BaseHtmlView;
use Joomla\CMS\Router\Route;
use Joomla\CMS\Toolbar\ToolbarHelper;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;

class HtmlView extends BaseHtmlView
{
	/**
	 * Is this an administrator page?
	 *
	 * @var   bool
	 */
	public $isAdmin = false;

	/**
	 * The MFA Methods available for this user
	 *
	 * @var   array
	 */
	public $methods = [];

	/**
	 * The return URL to use for all links and forms
	 *
	 * @var   string
	 */
	public $returnURL = null;

	/**
	 * Are there any active MFA Methods at all?
	 *
	 * @var   bool
	 */
	public $mfaActive = false;

	/**
	 * Which Method has the default record?
	 *
	 * @var   string
	 */
	public $defaultMethod = '';

	/**
	 * The user object used to display this page
	 *
	 * @var   User
	 */
	public $user = null;

	/**
	 * @inheritDoc
	 */
	function display($tpl = null)
	{
		$app = Factory::getApplication();

		if (empty($this->user))
		{
			$this->user = Factory::getApplication()->getIdentity()
				?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);
		}

		/** @var MethodsModel $model */
		$model = $this->getModel();

		if ($this->getLayout() != 'firsttime')
		{
			$this->setLayout('default');
		}

		$this->methods = $model->getMethods($this->user);
		$this->isAdmin = $app->isClient('administrator');
		$activeRecords = 0;

		foreach ($this->methods as $methodName => $method)
		{
			$methodActiveRecords = count($method['active']);

			if (!$methodActiveRecords)
			{
				continue;
			}

			$activeRecords   += $methodActiveRecords;
			$this->mfaActive = true;

			foreach ($method['active'] as $record)
			{
				if ($record->default)
				{
					$this->defaultMethod = $methodName;

					break;
				}
			}
		}

		// If there are no backup codes yet we should create new ones
		/** @var BackupcodesModel $model */
		$model       = $this->getModel('backupcodes');
		$backupCodes = $model->getBackupCodes($this->user);

		if ($activeRecords && empty($backupCodes))
		{
			$model->regenerateBackupCodes($this->user);
		}

		$backupCodesRecord = $model->getBackupCodesRecord($this->user);

		if (!is_null($backupCodesRecord))
		{
			$this->methods['backupcodes'] = new MethodDescriptor([
				'name'       => 'backupcodes',
				'display'    => Text::_('COM_LOGINGUARD_LBL_BACKUPCODES'),
				'shortinfo'  => Text::_('COM_LOGINGUARD_LBL_BACKUPCODES_DESCRIPTION'),
				'image'      => 'media/com_loginguard/images/emergency.svg',
				'canDisable' => false,
				'active'     => [$backupCodesRecord],
			]);
		}

		// Include CSS
		$this->document->getWebAssetManager()
		               ->useStyle('com_loginguard.methods');

		if (ComponentHelper::getParams('com_loginguard')->get('dark_mode') != 0)
		{
			$this->document->getWebAssetManager()
			               ->useStyle('com_loginguard.dark');
		}

		// Back-end: always show a title in the 'title' module position, not in the page body
		if ($this->isAdmin)
		{
			ToolbarHelper::title(Text::_('COM_LOGINGUARD_HEAD_LIST_PAGE'), 'loginguard');
			$this->title = '';

			ToolbarHelper::back('JTOOLBAR_BACK', Route::_('index.php?option=com_loginguard'));
		}

		if (ComponentHelper::getParams('com_loginguard')->get('dark_mode') != 0)
		{
			$this->document->getWebAssetManager()->useStyle('com_loginguard.dark');
		}

		// Display the view
		parent::display($tpl);

		MfaHelper::triggerEvent(
			new NotifyActionLog('onComLoginGuardViewMethodsAfterDisplay', [$this])
		);
	}
}