<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\User\Loginguard\Extension;

defined('_JEXEC') or die;

use Akeeba\Component\LoginGuard\Administrator\Helper\Mfa as MfaHelper;
use Akeeba\Component\LoginGuard\Administrator\Model\RemembermeModel;
use Exception;
use InvalidArgumentException;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Form\Form;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Router\Route;
use Joomla\CMS\Table\Menu;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Database\DatabaseDriver;
use Joomla\Database\ParameterType;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;
use Joomla\Registry\Registry;
use Joomla\Utilities\ArrayHelper;

class LoginGuard extends CMSPlugin implements SubscriberInterface
{
	/**
	 * Application object.
	 *
	 * @var    CMSApplication
	 * @since  6.0.0
	 */
	protected $app;

	/**
	 * Database object.
	 *
	 * @var    DatabaseDriver
	 * @since  6.0.0
	 */
	protected $db;

	/**
	 * Forbid registration of legacy (Joomla 3) event listeners.
	 *
	 * @var    boolean
	 * @since  6.0.0
	 *
	 * @deprecated
	 */
	protected $allowLegacyListeners = false;

	/**
	 * Automatically load language files.
	 *
	 * @var    bool
	 * @since  6.0.0
	 */
	protected $autoloadLanguage = true;

	/**
	 * @inheritDoc
	 */
	public static function getSubscribedEvents(): array
	{
		if (!ComponentHelper::isEnabled('com_loginguard'))
		{
			return [];
		}

		return [
			'onContentPrepareForm' => 'onContentPrepareForm',
			'onUserAfterLogin'     => 'onUserAfterLogin',
			'onUserAfterLogout'    => 'onUserAfterLogout',
			'onUserAfterDelete'    => 'onUserAfterDelete',
		];
	}

	/**
	 * Adds additional fields to the user editing form
	 *
	 * @param   Event   $event   The event we are handling
	 *
	 * @return  void
	 *
	 * @throws  Exception
	 */
	public function onContentPrepareForm(Event $event): void
	{
		/**
		 * @var  Form  $form The form we are handling in this event
		 * @var  mixed $data The data being assigned to the form
		 */
		[$form, $data] = $event->getArguments();
		$result   = $event->getArgument('result') ?: [];
		$result[] = true;

		$event->setArgument('result', $result);

		if (!($form instanceof Form))
		{
			throw new InvalidArgumentException('JERROR_NOT_A_FORM');
		}

		// Check we are manipulating a valid form.
		$name = $form->getName();

		if (!in_array($name, ['com_admin.profile', 'com_users.user', 'com_users.profile', 'com_users.registration']))
		{
			return;
		}

		$layout = $this->app->input->getCmd('layout', 'default');

		/**
		 * Joomla is kinda brain-dead. When we have a menu item to the Edit Profile page it does not push the layout
		 * into the Input (as opposed with option and view) so I have to go in and dig it out myself. Yikes!
		 */
		$itemId = $this->app->input->getInt('Itemid');

		if ($itemId)
		{
			try
			{
				/** @var Menu $menuItem */
				$menuItem = Factory::getContainer()->has(Menu::class)
					? Factory::getContainer()->get(Menu::class)
					: new Menu($this->db);
				$menuItem->load($itemId);
				/** @noinspection PhpUndefinedFieldInspection */
				$uri    = new Uri($menuItem->link);
				$layout = $uri->getVar('layout', $layout);
			}
			catch (Exception $e)
			{
				return;
			}
		}

		if (!$this->app->isClient('administrator') && !in_array($layout, ['edit', 'default']))
		{
			return;
		}

		// Get the user ID
		$id = null;

		if (is_array($data))
		{
			$id = $data['id'] ?? null;
		}
		elseif (is_object($data) && is_null($data) && ($data instanceof Registry))
		{
			$id = $data->get('id');
		}
		elseif (is_object($data) && !is_null($data))
		{
			$id = $data->id ?? null;
		}

		if (empty($id))
		{
			return;
		}

		$user = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($id);

		// Make sure the loaded user is the correct one
		if ($user->id != $id)
		{
			return;
		}

		// Make sure I am either editing myself OR I am a Super User AND I'm not editing another Super User
		if (!MfaHelper::canEditUser($user))
		{
			return;
		}

		// Add the fields to the form.
		Form::addFormPath(JPATH_PLUGINS . '/' . $this->_type . '/' . $this->_name . '/forms');

		// Special handling for profile overview page
		if ($layout === 'default')
		{
			$mfaMethods = MfaHelper::getUserMfaRecords($id);

			/**
			 * We cannot pass a boolean or integer; if it's false/0 Joomla! will display "No information entered". We
			 * cannot use a list field to display it in a human readable format, Joomla! just dumps the raw value if you
			 * use such a field. So all I can do is pass raw text. Um, whatever.
			 */
			$data->loginguard = [
				'hastfa' => (count($mfaMethods) > 0) ? Text::_('PLG_USER_LOGINGUARD_FIELD_HASTFA_ENABLED') : Text::_('PLG_USER_LOGINGUARD_FIELD_HASTFA_DISABLED'),
			];

			$form->loadFile('list', false);

			return;
		}

		// Profile edit page
		$form->loadFile('loginguard', false);

		return;
	}

	/**
	 * Runs after successful login of the user. Used to redirect the user to a page where they can set up their
	 * Multi-factor Authentication after logging in.
	 *
	 */
	public function onUserAfterLogin(Event $event)
	{
		/**
		 * @var  array $options Passed by Joomla. user: a User object; responseType: string, authentication response type.
		 */
		[$options] = $event->getArguments();

		// Make sure the option to redirect is set
		if (!$this->params->get('redirectonlogin', 1))
		{
			return;
		}

		// Make sure I have a valid user
		/** @var User $user */
		$user = $options['user'];

		if (!is_object($user) || !($user instanceof User))
		{
			return;
		}

		/**
		 * If the user already has MFA enabled and we need to show the Captive page we won't redirect them to the MFA
		 * setup page, of course.
		 *
		 * If the user has already asked us to not show him the MFA setup page we have to honour their wish.
		 */
		if (
			$this->needsCaptivePage($user, $options['responseType'])
			|| $this->hasDoNotShowAgainFlag($user)
		)
		{
			return;
		}

		// Get the redirection URL to the MFA setup page or custom redirection per plugin configuration
		$url = $this->params->get('redirecturl', null) ?:
			Route::_('index.php?option=com_loginguard&view=methods&layout=firsttime', false);

		// Prepare to redirect
		$this->app->getSession()->set('com_loginguard.postloginredirect', $url);
	}

	/**
	 * Fires after the user has logged out
	 *
	 * Used to remove the MFA Remember Me cookie from the browser.
	 *
	 * @param   Event   $event   The event we are handling
	 *
	 * @return  void
	 * @throws  Exception
	 */
	public function onUserAfterLogout(Event $event): void
	{
		/**
		 * @var   array|null $options
		 */
		[$options] = $event->getArguments();
		$result   = $event->getArgument('result') ?: [];
		$result[] = true;
		$event->setArgument('result', $result);

		// Is the Remember Me feature enabled?
		$allowRememberMe = ComponentHelper::getParams('com_loginguard')->get('allow_rememberme', 1);

		if ($allowRememberMe != 1)
		{
			return;
		}

		// Make sure I can get an non-empty username
		if (!is_array($options) || !array_key_exists('username', $options))
		{
			return;
		}

		$userName = $options['username'] ?? '';

		if (empty($userName))
		{
			return;
		}

		// Finally, remove the Remember Me cookie
		/** @var \Joomla\CMS\MVC\Factory\MVCFactoryInterface $factory */
		$factory = $this->app->bootComponent('com_loginguard')->getMVCFactory();
		/** @var RemembermeModel $rememberModel */
		$rememberModel = $factory->createModel('Rememberme', 'Administrator');
		$rememberModel->setUsername($userName)->removeCookie();

		return;
	}

	/**
	 * Remove all user profile information for the given user ID
	 *
	 * Method is called after user data is deleted from the database
	 *
	 * @param   Event   $event   The event we are handling
	 *
	 * @return  void
	 */
	public function onUserAfterDelete(Event $event): void
	{
		/**
		 * @var array  $user    Holds the user data
		 * @var bool   $success True if user was successfully stored in the database
		 * @var string $msg     Message
		 */
		[$user, $success, $msg] = $event->getArguments();
		$result = $event->getArgument('result') ?: [];

		if (!$success)
		{
			$result[] = false;
			$event->setArgument('result', $result);

			return;
		}

		$userId = ArrayHelper::getValue($user, 'id', 0, 'int');

		if (!$userId)
		{
			$result[] = true;
			$event->setArgument('result', $result);

			return;
		}

		$db = $this->db;

		// Delete user profile records
		$profileKey = 'loginguard.%';
		$query      = $db->getQuery(true)
			->delete($db->quoteName('#__user_profiles'))
			->where($db->quoteName('user_id') . ' = :userId')
			->where($db->quoteName('profile_key') . ' LIKE :profileKey')
			->bind(':userId', $userId, ParameterType::INTEGER)
			->bind(':profileKey', $profileKey, ParameterType::STRING);

		try
		{
			$db->setQuery($query)->execute();
		}
		catch (Exception $e)
		{
			// No sweat if it failed
		}

		// Delete LoginGuard records
		$query = $db->getQuery(true)
			->delete($db->qn('#__loginguard_tfa'))
			->where($db->quoteName('user_id') . ' = :userId')
			->bind(':userId', $userId, ParameterType::INTEGER);

		try
		{
			$db->setQuery($query)->execute();
		}
		catch (Exception $e)
		{
			// No sweat if it failed
		}

		$result[] = true;
		$event->setArgument('result', $result);
	}

	/**
	 * Does the current user need to complete 2FA authentication before allowed to access the site?
	 *
	 * @param   User          $user           The user object we are checking
	 * @param   string|null   $responseType   The login response type (optional)
	 *
	 * @return  bool
	 * @throws  Exception
	 */
	private function needsCaptivePage(User $user, ?string $responseType = null): bool
	{
		// Get the user's MFA records
		$records = MfaHelper::getUserMfaRecords($user->id);

		// No MFA Methods? Then we obviously don't need to display a Captive login page.
		if (count($records) < 1)
		{
			return false;
		}

		// Let's get a list of all currently active MFA Methods
		$mfaMethods = MfaHelper::getMfaMethods();

		// If not MFA Method is active we can't really display a Captive login page.
		if (empty($mfaMethods))
		{
			return false;
		}

		// Get a list of just the Method names
		$methodNames = [];

		foreach ($mfaMethods as $mfaMethod)
		{
			$methodNames[] = $mfaMethod['name'];
		}

		// Filter the records based on currently active MFA Methods
		foreach ($records as $record)
		{
			if (in_array($record->method, $methodNames))
			{
				// We found an active Method. Show the Captive page.
				return true;
			}
		}

		// No viable MFA Method found. We won't show the Captive page.
		return false;
	}

	/**
	 * Does the user have a "don't show this again" flag?
	 *
	 * @param   User   $user   The user to check
	 *
	 * @return  bool
	 */
	private function hasDoNotShowAgainFlag(User $user): bool
	{
		$profileKey = 'loginguard.dontshow';
		$db         = $this->db;
		$query      = $db->getQuery(true)
			->select($db->quoteName('profile_value'))
			->from($db->quoteName('#__user_profiles'))
			->where($db->quoteName('user_id') . ' = :userId')
			->where($db->quoteName('profile_key') . ' = :profileKey')
			->bind(':userId', $user->id, ParameterType::INTEGER)
			->bind(':profileKey', $profileKey, ParameterType::STRING);

		try
		{
			$result = $db->setQuery($query)->loadResult();
		}
		catch (Exception $e)
		{
			$result = 1;
		}

		return $result == 1;
	}
}