<?php
/**
 * @package   AkeebaLoginGuard
 * @copyright Copyright (c)2016-2022 Nicholas K. Dionysopoulos / Akeeba Ltd
 * @license   GNU General Public License version 3, or later
 */

namespace Joomla\Plugin\Loginguard\Webauthn\Extension;

// Prevent direct access
defined('_JEXEC') || die;

use Akeeba\Component\LoginGuard\Administrator\DataShape\CaptiveRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\DataShape\MethodDescriptor;
use Akeeba\Component\LoginGuard\Administrator\DataShape\SetupRenderOptions;
use Akeeba\Component\LoginGuard\Administrator\Event\GetInfo;
use Akeeba\Component\LoginGuard\Administrator\Event\SaveSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowCaptive;
use Akeeba\Component\LoginGuard\Administrator\Event\ShowSetup;
use Akeeba\Component\LoginGuard\Administrator\Event\Validate;
use Akeeba\Component\LoginGuard\Administrator\Table\TfaTable as TfaTable;
use Exception;
use Joomla\CMS\Application\AdministratorApplication;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Application\SiteApplication;
use Joomla\CMS\Factory;
use Joomla\CMS\Input\Input;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\Event\Event;
use Joomla\Event\SubscriberInterface;
use Joomla\Plugin\Loginguard\Webauthn\Helper\Credentials;
use RuntimeException;
use Webauthn\PublicKeyCredentialRequestOptions;

/**
 * Akeeba LoginGuard Plugin for Multi-factor Authentication Method "Webauthn Security Key"
 *
 * Uses W3C Web Authentication for Multi-factor Authentication.
 *
 * @since   3.1.0
 */
class Webauthn extends CMSPlugin implements SubscriberInterface
{
	/**
	 * The application object
	 *
	 * @var    CMSApplication|SiteApplication|AdministratorApplication
	 * @since  6.0.0
	 */
	protected $app;

	/**
	 * Auto-load the plugin's language files
	 *
	 * @var    boolean
	 * @since  5.0
	 */
	protected $autoloadLanguage = true;

	/**
	 * The MFA Method name handled by this plugin
	 *
	 * @var   string
	 * @since 3.1.0
	 */
	private $mfaMethodName = 'webauthn';

	/**
	 * Returns an array of events this subscriber will listen to.
	 *
	 * @return  array
	 *
	 * @since   6.0.0
	 */
	public static function getSubscribedEvents(): array
	{
		return [
			'onLoginGuardMfaGetMethod' => 'onLoginGuardMfaGetMethod',
			'onLoginGuardMfaCaptive'   => 'onLoginGuardMfaCaptive',
			'onLoginGuardMfaGetSetup'  => 'onLoginGuardMfaGetSetup',
			'onLoginGuardMfaSaveSetup' => 'onLoginGuardMfaSaveSetup',
			'onLoginGuardMfaValidate'  => 'onLoginGuardMfaValidate',
		];
	}

	/**
	 * Gets the identity of this MFA Method
	 *
	 * @param   GetInfo  $event
	 *
	 * @return  void
	 *
	 * @since   3.1.0
	 */
	public function onLoginGuardMfaGetMethod(GetInfo $event): void
	{
		$event->addResult(
			new MethodDescriptor([
				'name'               => $this->mfaMethodName,
				'display'            => Text::_('PLG_LOGINGUARD_WEBAUTHN_LBL_DISPLAYEDAS'),
				'shortinfo'          => Text::_('PLG_LOGINGUARD_WEBAUTHN_LBL_SHORTINFO'),
				'image'              => 'media/plg_loginguard_webauthn/images/final-webauthn-logo-webauthn-color.png',
				'allowMultiple'      => true,
				'help_url'           => $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Webauthn'),
				'allowEntryBatching' => true,
			])
		);
	}

	/**
	 * Returns the information which allows LoginGuard to render the MFA setup page. This is the page which allows the
	 * user to add or modify a MFA Method for their user account. If the record does not correspond to your plugin
	 * return an empty array.
	 *
	 * @param   ShowSetup  $event
	 *
	 * @return  void
	 * @throws  Exception
	 * @since   3.1.0
	 */
	public function onLoginGuardMfaGetSetup(ShowSetup $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			return;
		}

		// Get some values assuming that we are NOT setting up U2F (the key is already registered)
		$submitClass = '';
		$preMessage  = Text::_('PLG_LOGINGUARD_WEBAUTHN_LBL_CONFIGURED');
		$type        = 'input';
		$html        = '';
		$helpURL     = $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Webauthn');
		$hiddenData  = [];

		/**
		 * If there are no authenticators set up yet I need to show a different message and take a different action when
		 * my user clicks the submit button.
		 */
		if (!is_array($record->options) || empty($record->options) || !isset($record->options['credentialId']) || empty($record->options['credentialId']))
		{
			$document = $this->app->getDocument();
			$wam      = $document->getWebAssetManager();
			$wam->getRegistry()->addExtensionRegistryFile('plg_loginguard_webauthn');
			$wam->useScript('plg_loginguard_webauthn.webauthn');

			$layoutPath = PluginHelper::getLayoutPath('loginguard', 'webauthn', 'register');
			ob_start();
			include $layoutPath;
			$html = ob_get_clean();
			$type = 'custom';

			// Load JS translations
			Text::script('PLG_LOGINGUARD_WEBAUTHN_ERR_NOTAVAILABLE_HEAD');

			$document->addScriptOptions('com_loginguard.pagetype', 'setup', false);

			// Save the WebAuthn request to the session
			$user                    = Factory::getApplication()->getIdentity()
				?: Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById(0);
			$hiddenData['pkRequest'] = base64_encode(Credentials::requestAttestation($user));

			// Special button handling
			$submitClass = "loginguard_webauthn_setup";

			// Message to display
			$preMessage = Text::_('PLG_LOGINGUARD_WEBAUTHN_LBL_INSTRUCTIONS');
		}

		$event->addResult(
			new SetupRenderOptions([
				'default_title' => Text::_('PLG_LOGINGUARD_WEBAUTHN_LBL_DISPLAYEDAS'),
				'pre_message'   => $preMessage,
				'hidden_data'   => $hiddenData,
				'field_type'    => $type,
				'input_type'    => 'hidden',
				'html'          => $html,
				'show_submit'   => false,
				'submit_class'  => $submitClass,
				'help_url'      => $helpURL,
			])
		);
	}

	/**
	 * Parse the input from the MFA setup page and return the configuration information to be saved to the database. If
	 * the information is invalid throw a RuntimeException to signal the need to display the editor page again. The
	 * message of the exception will be displayed to the user. If the record does not correspond to your plugin return
	 * an empty array.
	 *
	 * @param   SaveSetup  $event
	 *
	 * @return  void
	 *
	 * @since   3.1.0
	 */
	public function onLoginGuardMfaSaveSetup(SaveSetup $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 * @var   Input    $input  The user input you are going to take into account.
		 */
		$record = $event['record'];
		$input  = $event['input'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			return;
		}

		$code                = $input->get('code', null, 'base64');
		$session             = $this->app->getSession();
		$registrationRequest = $session->get('plg_loginguard_webauthn.publicKeyCredentialCreationOptions', null);

		// If there was no registration request BUT there is a registration response throw an error
		if (empty($registrationRequest) && !empty($code))
		{
			throw new RuntimeException(Text::_('JERROR_ALERTNOAUTHOR'), 403);
		}

		// If there is no registration request (and there isn't a registration response) we are just saving the title.
		if (empty($registrationRequest))
		{
			$event->addResult($record->options);

			return;
		}

		// In any other case try to authorize the registration
		try
		{
			$publicKeyCredentialSource = Credentials::verifyAttestation($code);
		}
		catch (Exception $err)
		{
			throw new RuntimeException($err->getMessage(), 403);
		}
		finally
		{
			// Unset the request data from the session.
			$session->set('plg_loginguard_webauthn.publicKeyCredentialCreationOptions', null);
			$session->set('plg_loginguard_webauthn.registration_user_id', null);
		}

		// Return the configuration to be serialized
		$event->addResult([
			'credentialId' => base64_encode($publicKeyCredentialSource->getAttestedCredentialData()->getCredentialId()),
			'pubkeysource' => json_encode($publicKeyCredentialSource),
			'counter'      => 0,
		]);
	}

	/**
	 * Returns the information which allows LoginGuard to render the Captive MFA page. This is the page which appears
	 * right after you log in and asks you to validate your login with MFA.
	 *
	 * @param   ShowCaptive  $event
	 *
	 * @return  void
	 * @throws  Exception
	 * @since   3.1.0
	 */
	public function onLoginGuardMfaCaptive(ShowCaptive $event): void
	{
		/**
		 * @var   TfaTable $record The #__loginguard_tfa record currently selected by the user.
		 */
		$record = $event['record'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			return;
		}

		/**
		 * The following code looks stupid. An explanation is in order.
		 *
		 * What we normally want to do is save the authentication data returned by getAuthenticateData into the session.
		 * This is what is sent to the authenticator through the Javascript API and signed. The signature is posted back
		 * to the form as the "code" which is read by onLoginGuardMfaValidate. That Method will read the authentication
		 * data from the session and pass it along with the key registration data (from the database) and the
		 * authentication response (the "code" submitted in the form) to the WebAuthn library for validation.
		 *
		 * Validation will work as long as the challenge recorded in the encrypted AUTHENTICATION RESPONSE matches, upon
		 * decryption, the challenge recorded in the AUTHENTICATION DATA.
		 *
		 * I observed that for whatever stupid reason the browser was sometimes sending TWO requests to the server's
		 * Captive login page but only rendered the FIRST. This meant that the authentication data sent to the key had
		 * already been overwritten in the session by the "invisible" second request. As a result the challenge would
		 * not match and we'd get a validation error.
		 *
		 * The code below will attempt to read the authentication data from the session first. If it exists it will NOT
		 * try to replace it (technically it replaces it with a copy of the same data - same difference!). If nothing
		 * exists in the session, however, it WILL store the (random seeded) result of the getAuthenticateData Method.
		 * Therefore the first request to the Captive login page will store a new set of authentication data whereas the
		 * second, "invisible", request will just reuse the same data as the first request, fixing the observed issue in
		 * a way that doesn't compromise security.
		 *
		 * In case you are wondering, yes, the data is removed from the session in the onLoginGuardMfaValidate Method.
		 * In fact it's the first thing we do after reading it, preventing constant reuse of the same set of challenges.
		 *
		 * That was fun to debug - for "poke your eyes with a rusty fork" values of fun.
		 */

		$session          = $this->app->getSession();
		$pkOptionsEncoded = $session->get('plg_loginguard_webauthn.publicKeyCredentialRequestOptions', null);

		$force = $this->app->input->getInt('force', 0);

		try
		{
			if ($force)
			{
				throw new RuntimeException('Expected exception (good): force a new key request');
			}

			if (empty($pkOptionsEncoded))
			{
				throw new RuntimeException('Expected exception (good): we do not have a pending key request');
			}

			$serializedOptions = base64_decode($pkOptionsEncoded);
			$pkOptions         = unserialize($serializedOptions);

			if (!is_object($pkOptions) || empty($pkOptions) || !($pkOptions instanceof PublicKeyCredentialRequestOptions))
			{
				throw new RuntimeException('The pending key request is corrupt; a new one will be created');
			}

			$pkRequest = json_encode($pkOptions, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE);
		}
		catch (Exception $e)
		{
			$pkRequest = Credentials::requestAssertion($record->user_id);
		}

		try
		{
			/** @var CMSApplication $app */
			$app = Factory::getApplication();
			$app->getDocument()->addScriptOptions('com_loginguard.authData', base64_encode($pkRequest), false);
			$layoutPath = PluginHelper::getLayoutPath('loginguard', 'webauthn', 'validate');
			ob_start();
			include $layoutPath;
			$html = ob_get_clean();
		}
		catch (Exception $e)
		{
			return;
		}

		$document = $this->app->getDocument();
		$wam      = $document->getWebAssetManager();
		$wam->getRegistry()->addExtensionRegistryFile('plg_loginguard_webauthn');
		$wam->useScript('plg_loginguard_webauthn.webauthn');

		// Load JS translations
		Text::script('PLG_LOGINGUARD_WEBAUTHN_ERR_NOTAVAILABLE_HEAD');
		Text::script('PLG_LOGINGUARD_WEBAUTHN_ERR_NO_STORED_CREDENTIAL');

		$document->addScriptOptions('com_loginguard.pagetype', 'validate', false);

		$helpURL = $this->params->get('helpurl', 'https://github.com/akeeba/loginguard/wiki/Webauthn');

		$event->addResult(
			new CaptiveRenderOptions([
				// Custom HTML to display above the MFA form
				'pre_message'        => Text::_('PLG_LOGINGUARD_WEBAUTHN_LBL_INSTRUCTIONS'),
				// How to render the MFA code field. "input" (HTML input element) or "custom" (custom HTML)
				'field_type'         => 'custom',
				// The type attribute for the HTML input box. Typically "text" or "password". Use any HTML5 input type.
				'input_type'         => '',
				// Placeholder text for the HTML input box. Leave empty if you don't need it.
				'placeholder'        => '',
				// Label to show above the HTML input box. Leave empty if you don't need it.
				'label'              => '',
				// Custom HTML. Only used when field_type = custom.
				'html'               => $html,
				// Custom HTML to display below the MFA form
				'post_message'       => '',
				// Should I hide the submit button? Useful if you need to render your own buttons or use a Method which is meant to auto-submit upon doing a certain action.
				'hide_submit'        => true,
				// URL for help content
				'help_url'           => $helpURL,
				// Allow authentication against all entries of this MFA Method. Otherwise authentication takes place against a SPECIFIC entry at a time.
				'allowEntryBatching' => true,
			])
		);
	}

	/**
	 * Validates the Two Factor Authentication code submitted by the user in the Captive Multi-factor Authentication
	 * page. If the record does not correspond to your plugin return FALSE.
	 *
	 * @param   Validate  $event
	 *
	 * @return  void
	 * @since   3.1.0
	 */
	public function onLoginGuardMfaValidate(Validate $event): void
	{
		/**
		 * @var   TfaTable $record The MFA Method's record you're validatng against
		 * @var   User     $user   The user record
		 * @var   string   $code   The submitted code
		 */
		$record = $event['record'];
		$user   = $event['user'];
		$code   = $event['code'];

		// Make sure we are actually meant to handle this Method
		if ($record->method != $this->mfaMethodName)
		{
			$event->addResult(false);

			return;
		}

		// Double check the MFA Method is for the correct user
		if ($user->id != $record->user_id)
		{
			$event->addResult(false);

			return;
		}

		try
		{
			Credentials::verifyAssertion($code);
		}
		catch (Exception $e)
		{
			try
			{
				$this->app->enqueueMessage($e->getMessage(), 'error');
			}
			catch (Exception $e)
			{
			}

			$event->addResult(false);

			return;
		}

		$event->addResult(true);
	}
}
